# Copyright © 2024 Pathway

from __future__ import annotations

import pytest

import pathway as pw
from pathway import demo
from pathway.internals import Schema, api
from pathway.tests.utils import (
    DiffEntry,
    T,
    assert_key_entries_in_stream_consistent,
    assert_stream_equality,
    assert_stream_equality_wo_index,
    run,
)


def test_stream_success():
    class TimeColumnInputSchema(Schema):
        number: int
        parity: int

    value_functions = {
        "number": lambda x: x + 1,
        "parity": lambda x: (x + 1) % 2,
    }

    t = demo.generate_custom_stream(
        value_functions,
        schema=TimeColumnInputSchema,
        nb_rows=15,
        input_rate=15,
        autocommit_duration_ms=50,
    )

    gb = t.groupby(t.parity).reduce(t.parity, cnt=pw.reducers.count())

    list = []
    row: dict[str, api.Value]

    for i in [1, 2]:
        parity = i % 2
        row = {"cnt": 1, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i, True, row))
    for i in range(3, 16):
        parity = i % 2
        row = {"cnt": (i - 1) // 2, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i, False, row))
        row = {"cnt": (i + 1) // 2, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i, True, row))

    assert_key_entries_in_stream_consistent(list, gb)
    run()


def test_subscribe_consistency():
    class TimeColumnInputSchema(Schema):
        number: int
        parity: int
        number_div: int

    value_functions = {
        "number": lambda x: x + 1,
        "number_div": lambda x: (x + 1) // 2,
        "parity": lambda x: (x + 1) % 2,
    }

    t = demo.generate_custom_stream(
        value_functions,
        schema=TimeColumnInputSchema,
        nb_rows=15,
        input_rate=15,
        autocommit_duration_ms=100,
    )

    gb = t.groupby(t.parity).reduce(
        t.parity,
        cnt=pw.reducers.count(),
        max_number=pw.reducers.max(t.number),
        max_div=pw.reducers.max(t.number_div),
    )

    list = []
    row: dict[str, api.Value]

    for i in [1, 2]:
        parity = i % 2
        row = {"cnt": 1, "parity": parity, "max_number": i, "max_div": i // 2}
        list.append(DiffEntry.create(gb, {"parity": parity}, i, True, row))
    for i in range(3, 16):
        parity = i % 2
        row = {
            "cnt": (i - 1) // 2,
            "parity": parity,
            "max_number": i - 2,
            "max_div": (i - 2) // 2,
        }
        list.append(DiffEntry.create(gb, {"parity": parity}, i, False, row))
        row = {
            "cnt": (i + 1) // 2,
            "parity": parity,
            "max_number": i,
            "max_div": i // 2,
        }
        list.append(DiffEntry.create(gb, {"parity": parity}, i, True, row))

    assert_key_entries_in_stream_consistent(list, gb)
    run()


def test_stream_test_util_should_fail_q_none():
    class TimeColumnInputSchema(Schema):
        number: int
        parity: int

    value_functions = {
        "number": lambda x: x + 1,
        "parity": lambda x: (x + 1) % 2,
    }

    t = demo.generate_custom_stream(
        value_functions,
        schema=TimeColumnInputSchema,
        nb_rows=15,
        input_rate=15,
        autocommit_duration_ms=50,
    )

    gb = t.groupby(t.parity).reduce(t.parity, cnt=pw.reducers.count())

    list = []
    row: dict[str, api.Value]

    for i in [1, 2]:
        parity = i % 2
        row = {"cnt": 1, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i, True, row))
    for i in range(3, 7):
        parity = i % 2
        row = {"cnt": (i - 1) // 2, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i + 7, False, row))
        row = {"cnt": (i + 1) // 2, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i + 7, True, row))
    for i in range(7, 16):
        parity = i % 2
        row = {"cnt": (i - 1) // 2, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i - 4, False, row))
        row = {"cnt": (i + 1) // 2, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i - 4, True, row))

    assert_key_entries_in_stream_consistent(list, gb)
    with pytest.raises(AssertionError):
        run()


def test_stream_test_util_should_fail_empty_final_state():
    class TimeColumnInputSchema(Schema):
        number: int
        parity: int

    value_functions = {
        "number": lambda x: x + 1,
        "parity": lambda x: (x + 1) % 2,
    }

    t = demo.generate_custom_stream(
        value_functions,
        schema=TimeColumnInputSchema,
        nb_rows=15,
        input_rate=15,
        autocommit_duration_ms=50,
    )

    gb = t.groupby(t.parity).reduce(t.parity, cnt=pw.reducers.count())

    list = []
    row: dict[str, api.Value]

    for i in [1, 2]:
        parity = i % 2
        row = {"cnt": 1, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i, True, row))
    for i in range(3, 18):
        parity = i % 2
        row = {"cnt": (i - 1) // 2, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i, False, row))
        row = {"cnt": (i + 1) // 2, "parity": parity}
        list.append(DiffEntry.create(gb, {"parity": parity}, i, True, row))

    assert_key_entries_in_stream_consistent(list, gb)
    with pytest.raises(AssertionError):
        run()


def test_assert_stream_equality():
    t = T(
        """
      | a | __time__ | __diff__
    9 | 0 | 2        |    1
    7 | 2 | 4        |    1
    8 | 1 | 4        |    1
    6 | 3 | 6        |    1
    7 | 2 | 6        |   -1
    6 | 4 | 8        |    1
    5 | 4 | 8        |    1
    6 | 3 | 8        |   -1
    """
    )
    expected = T(
        """
      | a | __time__ | __diff__
    9 | 0 | 2        |    1
    8 | 1 | 4        |    1
    7 | 2 | 4        |    1
    6 | 3 | 6        |    1
    7 | 2 | 6        |   -1
    5 | 4 | 8        |    1
    6 | 3 | 8        |   -1
    6 | 4 | 8        |    1
    """
    )
    assert_stream_equality(t, expected)


def test_assert_table_revisions_equality_with_id():
    t = T(
        """
    a | __time__ | __diff__
    0 |    2     |    1
    1 |    4     |    1
    2 |    4     |    1
    3 |    6     |    1
    4 |    8     |    1
    4 |    8     |    1
    """
    )
    expected = T(
        """
    a | __time__ | __diff__
    0 |    2     |    1
    2 |    4     |    1
    1 |    4     |    1
    3 |    6     |    1
    4 |    8     |    1
    4 |    8     |    1
    """
    )
    assert_stream_equality_wo_index(t, expected)


def test_assert_table_revisions_equality_with_id_multiple_workers():
    t = T(
        """
      | a | __time__ | __diff__ | __shard__
    9 | 0 |    2     |    1     |     0
    8 | 1 |    4     |    1     |     0
    7 | 2 |    4     |    1     |     1
    2 | 8 |    4     |    1     |     2
    6 | 3 |    6     |    1     |     3
    7 | 2 |    6     |   -1     |     1
    5 | 4 |    8     |    1     |     0
    6 | 3 |    8     |   -1     |     3
    6 | 4 |    8     |    1     |     3
    2 | 8 |    8     |   -1     |     2
    """
    )
    expected = T(
        """
      | a | __time__ | __diff__
    9 | 0 |    2     |    1
    8 | 1 |    4     |    1
    7 | 2 |    4     |    1
    2 | 8 |    4     |    1
    6 | 3 |    6     |    1
    7 | 2 |    6     |   -1
    5 | 4 |    8     |    1
    6 | 3 |    8     |   -1
    6 | 4 |    8     |    1
    2 | 8 |    8     |   -1
    """
    )
    assert_stream_equality(t, expected)


def test_raises_when_not_equal_1():
    t = T(
        """
      | a | __time__ | __diff__
    9 | 0 |    2     |    1
    8 | 1 |    4     |    1
    7 | 2 |    4     |    1
    """
    )
    expected = T(
        """
      | a | __time__ | __diff__
    9 | 0 |    2     |    1
    """
    )
    with pytest.raises(AssertionError):
        assert_stream_equality(t, expected)


def test_raises_when_not_equal_2():
    t = T(
        """
      | a | __time__ | __diff__
    9 | 0 |    2     |    1
    8 | 1 |    4     |    1
    7 | 2 |    4     |    1
    """
    )
    expected = T(
        """
      | a | __time__ | __diff__
    9 | 0 |    2     |    1
    7 | 2 |    4     |    1
    """
    )
    with pytest.raises(AssertionError):
        assert_stream_equality(t, expected)


def test_raises_when_schemas_dont_match():
    t = T(
        """
      | a | __time__ | __diff__
    9 | 0 |    2     |    1
    8 | 1 |    4     |    1
    7 | 2 |    4     |    1
    """
    )
    expected = T(
        """
      | b | __time__ | __diff__
    9 | 0 |    2     |    1
    8 | 1 |    4     |    1
    7 | 2 |    4     |    1
    """
    )
    with pytest.raises(RuntimeError):
        assert_stream_equality(t, expected)


def test_compute_and_print_update_stream(capsys):
    table_def = """
      | a | __time__ | __diff__
    9 | 0 |    2     |    1
    8 | 1 |    4     |    1
    7 | 2 |    4     |    1
    8 | 1 |    6     |   -1
    8 | 2 |    6     |    1
    """
    expected = """
a | __time__ | __diff__
0 | 2        | 1
1 | 4        | 1
2 | 4        | 1
1 | 6        | -1
2 | 6        | 1
    """
    t = T(table_def)
    pw.debug.compute_and_print_update_stream(t, include_id=False)
    caputed = capsys.readouterr()
    assert caputed.out.strip() == expected.strip()


def test_compute_and_print(capsys):
    table_def = """
      | a | __time__ | __diff__
    9 | 0 |    2     |    1
    8 | 1 |    4     |    1
    7 | 2 |    4     |    1
    8 | 1 |    6     |   -1
    8 | 2 |    6     |    1
    """
    expected = """
a
0
2
2
    """
    t = T(table_def)
    pw.debug.compute_and_print(t, include_id=False)
    caputed = capsys.readouterr()
    assert caputed.out.strip() == expected.strip()
